

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.lines as mlines
import datastock as ds


# ################################################
# ################################################
#           Main
# ################################################


def main(
    det_size=None,
    ap_size=None,
    focal=None,
    pitch=None,
    det_nb=None,
    dist=None,
):
    """ Plot a 2d represntation of a pinhole camera

    Simple planar pinhole camera model, with parameterized:
        - det_size: detector size
        - ap_size: aperture size
        - focal: focal length
        - pitch: distance between sensors center to center
        - det_nb: nb of sensors, has to be an odd integer
        - dist: dist up to which plotting should be done

    """

    # ---------------
    # check inputs
    # ---------------

    (
        det_size,
        ap_size,
        focal,
        pitch,
        det_nb,
        dist,
    ) = _check(
        det_size=det_size,
        ap_size=ap_size,
        focal=focal,
        pitch=pitch,
        det_nb=det_nb,
        dist=dist,
    )

    # ---------------
    # compute
    # ---------------

    dout = _compute(
        det_size=det_size,
        ap_size=ap_size,
        focal=focal,
        pitch=pitch,
        det_nb=det_nb,
        dist=dist,
    )

    # ---------------
    # add text
    # ---------------

    _add_text(dout)

    # ---------------
    # plot
    # ---------------

    _plot(dout)

    return dout


# ################################################
# ################################################
#           Check
# ################################################


def _check(
    det_size=None,
    ap_size=None,
    focal=None,
    pitch=None,
    det_nb=None,
    dist=None,
):

    # -------------
    # sizes
    # -------------

    # det_size
    det_size = float(ds._generic_check._check_var(
        det_size, 'det_size',
        types=(int, float),
        sign='>0',
        default=0.005,
    ))

    # ap_size
    ap_size = float(ds._generic_check._check_var(
        ap_size, 'ap_size',
        types=(int, float),
        sign='>0',
        default=0.005,
    ))

    # -------------
    # focal
    # -------------

    focal = float(ds._generic_check._check_var(
        focal, 'focal',
        types=(int, float),
        sign='>0',
        default=0.20,
    ))

    # -------------
    # pitch
    # -------------

    pitch = float(ds._generic_check._check_var(
        pitch, 'pitch',
        types=(int, float),
        sign='>0',
        default=0.02,
    ))

    # -------------
    # det_nb
    # -------------

    det_nb = ds._generic_check._check_var(
        det_nb, 'det_nb',
        types=int,
        sign='>0',
        default=3,
    )

    if det_nb % 2 == 0:
        det_nb = det_nb + 1

    # -------------
    # dist
    # -------------

    dist = float(ds._generic_check._check_var(
        dist, 'dist',
        types=(int, float),
        sign='>0',
        default=2.,
    ))

    return (
        det_size,
        ap_size,
        focal,
        pitch,
        det_nb,
        dist,
    )


# ################################################
# ################################################
#           Compute
# ################################################


def _compute(
    det_size=None,
    ap_size=None,
    focal=None,
    pitch=None,
    det_nb=None,
    dist=None,
):

    # ---------------
    # prepare & outputs
    # ---------------

    dist_plot = dist

    alpha = np.arctan(pitch / focal)
    beta = 2. * np.arctan((det_size + ap_size) / (2. * focal))
    R = (beta - alpha) / alpha

    # ---------------
    # sensors
    # ---------------

    cy = pitch * np.linspace(-1, 1, det_nb)

    y_up = cy + det_size * 0.5
    y_low = cy - det_size * 0.5

    x_plot = np.zeros((det_nb*3))
    y_plot = np.array([y_low, y_up, np.full(cy.shape, np.nan)]).T.ravel()

    # ---------------
    # aperture
    # ---------------

    apx = focal
    apy_up = 0.5 * ap_size
    apy_low = -0.5 * ap_size

    apx_plot = np.full((5,), apx)
    apy_plot = np.array([y_low[0], apy_low, np.nan, apy_up, y_up[-1]])

    # ---------------
    # LOS
    # ---------------

    vectx = apx
    vecty = -cy
    vectn = np.sqrt(vectx**2 + vecty**2)
    vectx = vectx / vectn
    vecty = vecty / vectn

    kplot = dist_plot / vectx

    los_x = np.r_[0, dist_plot]
    los_y = (
        cy[:, None]
        + kplot[:, None] * vecty[:, None] * np.r_[0, 1][None, :]
    )

    # ---------------
    # FOS
    # ---------------

    fov_x = np.r_[0, dist_plot]

    # ------
    # inner

    # vect low
    vlow_x = apx
    vlow_y = -0.5*ap_size - y_low
    vlown = np.sqrt(vlow_x**2 + vlow_y**2)
    vlow_x, vlow_y = vlow_x/vlown, vlow_y/vlown

    # vect up
    vup_x = apx
    vup_y = 0.5*ap_size - y_up
    vupn = np.sqrt(vup_x**2 + vup_y**2)
    vup_x, vup_y = vup_x/vupn, vup_y/vupn

    # y
    klow = dist_plot / vlow_x
    kup = dist_plot / vup_x
    fov_in_y_low = (
        y_low[:, None]
        + klow[:, None] * vlow_y[:, None] * np.r_[0, 1][None, :]
    )
    fov_in_y_up = (
        y_up[:, None]
        + kup[:, None] * vup_y[:, None] * np.r_[0, 1][None, :]
    )

    # ------
    # outer

    # vect low
    vlow_x = apx
    vlow_y = 0.5*ap_size - y_low
    vlown = np.sqrt(vlow_x**2 + vlow_y**2)
    vlow_x, vlow_y = vlow_x/vlown, vlow_y/vlown

    # vect up
    vup_x = apx
    vup_y = -0.5*ap_size - y_up
    vupn = np.sqrt(vup_x**2 + vup_y**2)
    vup_x, vup_y = vup_x/vupn, vup_y/vupn

    # y
    klow = dist_plot / vlow_x
    kup = dist_plot / vup_x
    fov_out_y_low = (
        y_low[:, None]
        + klow[:, None] * vlow_y[:, None] * np.r_[0, 1][None, :]
    )
    fov_out_y_up = (
        y_up[:, None]
        + kup[:, None] * vup_y[:, None] * np.r_[0, 1][None, :]
    )

    # ---------------
    # dout
    # ---------------

    dout = {
        'inputs': {
            'pitch': pitch,
            'focal': focal,
            'det_size': det_size,
            'ap_size': ap_size,
        },
        'outputs': {
            'alpha': alpha,
            'beta': beta,
            'R': R,
        },
        'sensors': {
            'cy': cy,
            'y_low': y_low,
            'y_up': y_up,
            'x_plot': x_plot,
            'y_plot': y_plot,
        },
        'aperture': {
            'cx': apx,
            'y_low': apy_low,
            'y_up': apy_up,
            'x_plot': apx_plot,
            'y_plot': apy_plot,
        },
        'LOS': {
            'vectx': vectx,
            'vecty': vecty,
            'x': los_x,
            'y': los_y,
        },
        'FOV': {
            'x': fov_x,
            'inner': {
                'y_low': fov_in_y_low,
                'y_up': fov_in_y_up,
            },
            'outer': {
                'y_low': fov_out_y_low,
                'y_up': fov_out_y_up,
            },
        },
    }

    return dout


# ################################################
# ################################################
#           Add text
# ################################################


def _add_text(dout):

    # -------------
    # alpha
    # -------------

    alpha = (
        r"$\alpha = \arctan\left(\frac{P}{F}\right)$"
    )

    # -------------
    # beta
    # -------------

    beta = (
        r"$\beta = 2\arctan\left(\frac{W_A + W_D}{2F}\right)$"
    )

    # -------------
    # overlap rate
    # -------------

    overlap = (
        "Overap rate:  "
        + r"$R = \frac{\beta - \alpha}{\alpha}$" + "\n   "
        + r"$\approx \frac{W_A + W_D - P}{F} \times \frac{F}{P}$" + "\n   "
        + r"$\approx \frac{W_A + W_D - P}{P}$" + "\n\nOr:   "
        + r"$W_A \approx P(1 + R) - W_D$"
    )

    # -------------
    # inputs
    # -------------

    inputs = (
        "Inputs:\n   "
        + r"$P = $" + f"{dout['inputs']['pitch']} m\n   "
        + r"$F = $" + f"{dout['inputs']['focal']} m\n   "
        + r"$W_D = $" + f"{dout['inputs']['det_size']} m\n   "
        + r"$W_A = $" + f"{dout['inputs']['ap_size']} m"
    )

    # -------------
    # outputs
    # -------------

    alpha_str = round(dout['outputs']['alpha']*180/np.pi, ndigits=1)
    beta_str = round(dout['outputs']['beta']*180/np.pi, ndigits=1)
    R_str = round(dout['outputs']['R']*100, ndigits=1)

    outputs = (
        "Outputs:\n   "
        + r"$\alpha = $" + f"{alpha_str} deg\n   "
        + r"$\beta = $" + f"{beta_str} deg\n   "
        + r"$R = $" + f"{R_str} %"
    )

    # -------------
    # Recommendation
    # -------------

    WA = 1.1*dout['inputs']['pitch'] - dout['inputs']['det_size']

    R10 = (
        "To get R = 10 %:\n    "
        + r"$W_A \approx 1.1P - W_D = $" + f"{WA:3.1e} m"
    )

    # -------------
    # store
    # -------------

    dout['text'] = {
        'alpha': {
            'str': alpha,
            'pos': (0.10, 0.40),
        },
        'beta': {
            'str': beta,
            'pos': (0.25, 0.40),
        },
        'overlap': {
            'str': overlap,
            'pos': (0.10, 0.30),
        },
        'inputs': {
            'str': inputs,
            'pos': (0.60, 0.40),
        },
        'outputs': {
            'str': outputs,
            'pos': (0.80, 0.40),
        },
        'R10': {
            'str': R10,
            'pos': (0.60, 0.15),
        },
    }

    return


# ################################################
# ################################################
#           Plot
# ################################################


def _plot(dout):

    # ----------------
    # prepare data
    # ----------------

    dmargin = {
        'bottom': 0.1, 'top': 0.9,
        'left': 0.1, 'right': 0.9,
        'wspace': 0.1, 'hspace': 0.1,
    }

    # ----------------
    # prepare figure
    # ----------------

    # figure
    fig = plt.figure(figsize=(12, 8))
    gs = gridspec.GridSpec(ncols=1, nrows=2, **dmargin)

    # axes
    ax = fig.add_subplot(gs[0, :], aspect='equal', adjustable='datalim')
    ax.set_xlabel('X (m)', size=12, fontweight='bold')
    ax.set_ylabel('Y (m)', size=12, fontweight='bold')

    # ------------------------
    # plot sensors & aperture
    # ------------------------

    for k0 in ['sensors', 'aperture']:
        ax.plot(
            dout[k0]['x_plot'],
            dout[k0]['y_plot'],
            c='k',
            lw=5,
        )

    # ------------------------
    # plot LOS and FOV
    # ------------------------

    for ii in range(dout['sensors']['cy'].size):

        # LOS
        l0, = ax.plot(
            dout['LOS']['x'],
            dout['LOS']['y'][ii],
            ls='-',
            lw=1,
        )

        # FOV inner
        ax.fill_between(
            dout['FOV']['x'],
            dout['FOV']['inner']['y_low'][ii],
            dout['FOV']['inner']['y_up'][ii],
            fc=l0.get_color(),
            alpha=0.6,
        )

        # FOV outer
        ax.fill_between(
            dout['FOV']['x'],
            dout['FOV']['outer']['y_low'][ii],
            dout['FOV']['outer']['y_up'][ii],
            fc=l0.get_color(),
            alpha=0.2,
        )

    # ----------------
    # text
    # ----------------

    for k0, v0 in dout['text'].items():

        ax.text(
            v0['pos'][0],
            v0['pos'][1],
            v0['str'],
            size=14,
            verticalalignment='top',
            horizontalalignment='left',
            transform=fig.transFigure,
        )

    # figure
    fig.add_artist(mlines.Line2D(
        [0.5, 0.5],
        [0.1, 0.4],
        linewidth=2,
        color='k',
    ))

    return





import numpy as np
import datastock as ds


# ########################################################
# ########################################################
#              DEFAULTS
# ########################################################


_DANGLES = {
    'angle0': {
        'nb': 101,
        'lim': np.r_[-1, 1] * 90,
        'units': 'deg',
    },
    'angle1': {
        'nb': 101,
        'lim': np.r_[-1, 1] * 90,
        'units': 'deg',
    },
}


# ########################################################
# ########################################################
#              Main
# ########################################################


def main(
    coll=None,
    key=None,
    cent=None,
    nin=None,
    e0=None,
    e1=None,
    angle0=None,
    angle1=None,
    config=None,
    strict=None,
    # optional naming
    key_angle0=None,
    key_angle1=None,
    ref_angle0=None,
    ref_angle1=None,
    units_angles=None,
):

    # -------------
    # check inputs
    # -------------

    key, cent, nin, e0, e1, dangles, strict = _check(
        coll=coll,
        key=key,
        cent=cent,
        nin=nin,
        e0=e0,
        e1=e1,
        angle0=angle0,
        angle1=angle1,
        strict=strict,
        # optional naming
        key_angle0=key_angle0,
        key_angle1=key_angle1,
        ref_angle0=ref_angle0,
        ref_angle1=ref_angle1,
        units_angles=units_angles,
    )

    # -------------
    # prepare
    # -------------

    # -----------
    # angles ref

    for k0, v0 in dangles.items():
        if v0['ref'] not in coll.dref.keys():
            coll.add_ref(v0['ref'], size=v0['data'].size)

    # ------------
    # angles data

    for k0, v0 in dangles.items():
        if v0['key'] not in coll.dref.keys():
            coll.add_data(**v0)

    # -------------
    # unit vectors
    # -------------

    # angles, full shape
    angle0f = dangles['angle0']['data'][:, None]
    if dangles['angle0']['units'] == 'deg':
        angle0f = angle0f * np.pi/180.

    angle1f = dangles['angle1']['data'][None, :]
    if dangles['angle1']['units'] == 'deg':
        angle1f = angle1f * np.pi/180.

    # unit vectors
    vx = (
        np.cos(angle1f)
        * (np.cos(angle0f) * (nin[0]) + np.sin(angle0f) * e0[0])
        + np.sin(angle1f) * e1[0]
    )
    vy = (
        np.cos(angle1f)
        * (np.cos(angle0f) * (nin[1]) + np.sin(angle0f) * e0[1])
        + np.sin(angle1f) * e1[1]
    )
    vz = (
        np.cos(angle1f)
        * (np.cos(angle0f) * (nin[2]) + np.sin(angle0f) * e0[2])
        + np.sin(angle1f) * e1[2]
    )

    # -------------
    # compute
    # -------------

    coll.add_rays(
        key,
        start_x=np.full(vx.shape, cent[0]),
        start_y=np.full(vx.shape, cent[1]),
        start_z=np.full(vx.shape, cent[2]),
        # vectors
        vect_x=vx,
        vect_y=vy,
        vect_z=vz,
        # ref
        ref=(dangles['angle0']['ref'], dangles['angle1']['ref']),
        # config
        config=config,
        strict=strict,
    )

    return


# ########################################################
# ########################################################
#              Check
# ########################################################


def _check(
    coll=None,
    key=None,
    cent=None,
    nin=None,
    e0=None,
    e1=None,
    angle0=None,
    angle1=None,
    strict=None,
    # optional naming
    key_angle0=None,
    key_angle1=None,
    ref_angle0=None,
    ref_angle1=None,
    units_angles=None,
):

    # --------------
    # key
    # --------------

    wrays = coll._which_rays
    lout = list(coll.dobj.get(wrays, {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        default='ptcam',
        excluded=lout,
    )

    # --------------
    # cent
    # --------------

    cent = ds._generic_check._check_flat1darray(
        cent, 'cent',
        dtype=float,
        size=3,
    )

    # --------------
    # unit vectors
    # --------------

    nin, e0, e1 = ds._generic_check._check_vectbasis(
        e0=nin,
        e1=e0,
        e2=e1,
        dim=3,
        direct=False,
    )

    # --------------
    # angle0, angle1 angles
    # --------------

    # angle0
    dangles = {
        'angle0': _check_angles(
            coll=coll,
            key=key,
            ang=angle0,
            ang_name='angle0',
            ang_key=key_angle0,
            ang_ref=ref_angle0,
        ),
        'angle1': _check_angles(
            coll=coll,
            key=key,
            ang=angle1,
            ang_name='angle1',
            ang_key=key_angle1,
            ang_ref=ref_angle1,
        ),
    }

    # --------------
    # options
    # --------------

    # strict
    strict = ds._generic_check._check_var(
        strict, 'strict',
        types=bool,
        default=False,
    )

    return key, cent, nin, e0, e1, dangles, strict


def _check_angles(
    coll=None,
    key=None,
    ang=None,
    ang_name=None,
    ang_key=None,
    ang_ref=None,
    ang_units=None,
):

    # --------------
    # existing key
    # --------------

    if isinstance(ang, str):
        lok = [
            k0 for k0, v0 in coll.ddata.items()
            if v0['monot'] == (True,)
        ]
        ang = ds._generic_check._check_var(
            ang, ang_name,
            types=str,
            allowed=lok,
        )

        ang_key = ang
        ang_data = coll.ddata[ang_key]['data']
        ang_ref = coll.ddata[ang_key]['ref'][0]
        ang_units = coll.ddata[ang_key]['units']

    # --------------
    # int, float, array
    # --------------

    else:

        if ang is None:
            ang = _DANGLES[ang_name]['nb']

        if isinstance(ang, (float, int)):
            ang = np.linspace(
                _DANGLES[ang_name]['lim'][0],
                _DANGLES[ang_name]['lim'][1],
                int(ang),
            )
            ang_units = _DANGLES[ang_name]['units']

        # --------------
        # flat unique array

        ang_data = ds._generic_check._check_flat1darray(
            ang, ang_name,
            dtype=float,
            unique=True,
        )

        # -------
        # keys

        lout = list(coll.ddata.keys())
        ang_key = ds._generic_check._check_var(
            ang_key, ang_name,
            types=str,
            default=f"{key}_{ang_name}",
            excluded=lout,
        )

        # -------
        # ref

        lout = list(coll.dref.keys())
        ang_ref = ds._generic_check._check_var(
            ang_ref, f"{ang_name}",
            types=str,
            default=f"{key}_n{ang_name}",
            excluded=lout,
        )

        # -------
        # units

        ang_units = ds._generic_check._check_var(
            ang_units, "units_angles",
            types=str,
            default="deg",
            allowed=['deg', 'rad'],
        )

    return {
        'key': ang_key,
        'data': ang_data,
        'ref': ang_ref,
        'units': ang_units,
    }





# Standard
import numpy as np


# tofu-specific
import tofu.physics_tools as tfpt


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module():
    pass


def teardown_module():
    pass


#######################################################
#
#     checking Distributions
#
#######################################################


class Test00_Geometry():

    @classmethod
    def setup_class(cls):
        pass

    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_pinhole_camera_estimator(self):
        dout = tfpt.geometry.pinhole_camera_estimator()
        assert isinstance(dout, dict)
